from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, field
from importlib.util import find_spec
from typing import Any, AsyncGenerator, List, Literal, Optional, Union, overload

from pydantic import BaseModel, field_validator

from ..models import Message, Resource


class ModelConfig(BaseModel):
    loader: str = ""
    """所使用加载器的名称，位于 llm 文件夹下，loader 开头必须大写"""

    template: Optional[str] = None
    """使用的人设模板名称"""
    template_mode: Literal["system", "user"] = "system"
    """模板嵌入模式: `system` 为嵌入到系统提示; `user` 为嵌入到用户提示中"""

    max_tokens: int = 4096
    """最大回复 Tokens """
    temperature: float = 0.75
    """模型的温度系数"""
    top_p: float = 0.95
    """模型的 top_p 系数"""
    top_k: float = 3
    """模型的 top_k 系数"""
    frequency_penalty: Optional[float] = None
    """模型的频率惩罚"""
    presence_penalty: Optional[float] = None
    """模型的存在惩罚"""
    repetition_penalty: Optional[float] = None
    """模型的重复惩罚"""
    stream: bool = False
    """是否使用流式输出"""
    online_search: bool = False
    """是否启用联网搜索（原生实现）"""
    function_call: bool = False
    """是否启用工具调用"""
    content_security: bool = False
    """是否启用内容安全"""

    model_path: str = ""
    """本地模型路径"""
    adapter_path: str = ""
    """基于 model_path 的微调模型或适配器路径"""

    model_name: str = ""
    """所要使用模型的名称"""
    api_key: str = ""
    """在线服务的 API KEY"""
    api_secret: str = ""
    """在线服务的 api secret """
    api_host: str = ""
    """自定义 API 地址"""

    extra_body: Optional[dict] = None
    """OpenAI 的 extra_body"""
    enable_thinking: Optional[bool] = None
    """Dashscope 的 enable_thinking"""
    thinking_budget: Optional[int] = None
    """Dashscope 的 thinking_budget"""

    multimodal: bool = False
    """是否为（或启用）多模态模型"""
    modalities: List[Literal["text", "audio", "image"]] = ["text"]
    """生成模态"""
    audio: Optional[Any] = None
    """多模态音频参数"""

    @field_validator("loader")
    @classmethod
    def check_model_loader(cls, loader) -> str:
        if not loader:
            raise ValueError("loader is required")

        # Check if the specified loader exists
        module_path = f"muicebot.llm.{loader}"

        # 使用 find_spec 仅检测模块是否存在，不实际导入
        if find_spec(module_path) is None:
            raise ValueError(f"指定的模型加载器 '{loader}' 不存在于 llm 目录中")

        return loader


class BasicModel(metaclass=ABCMeta):
    """
    模型基类，所有模型加载器都必须继承于该类

    推荐使用该基类中定义的方法构建模型加载器类，但无论如何都必须实现 `ask` 方法
    """

    def __init__(self, model_config: ModelConfig) -> None:
        """
        统一在此处声明变量
        """
        self.config = model_config
        """模型配置"""
        self.is_running = False
        """模型状态"""
        self._total_tokens = -1
        """本次总请求（包括工具调用）使用的总token数。当此值设为-1时，表明此模型加载器不支持该功能"""

    def _require(self, *require_fields: str):
        """
        通用校验方法：检查指定的配置项是否存在，不存在则抛出错误

        :param require_fields: 需要检查的字段名称（字符串）
        """
        missing_fields = [field for field in require_fields if not getattr(self.config, field, None)]
        if missing_fields:
            raise ValueError(f"对于 {self.config.loader} 以下配置是必需的: {', '.join(missing_fields)}")

    def _build_messages(self, request: "ModelRequest") -> list:
        """
        构建对话上下文历史的函数
        """
        raise NotImplementedError

    def load(self) -> bool:
        """
        加载模型（通常是耗时操作，在线模型如无需校验可直接返回 true）

        :return: 是否加载成功
        """
        self.is_running = True
        return True

    async def _ask_sync(self, messages: list) -> "ModelCompletions":
        """
        同步模型调用
        """
        raise NotImplementedError

    def _ask_stream(self, messages: list) -> AsyncGenerator["ModelStreamCompletions", None]:
        """
        流式输出
        """
        raise NotImplementedError

    @overload
    async def ask(self, request: "ModelRequest", *, stream: Literal[False] = False) -> "ModelCompletions": ...

    @overload
    async def ask(
        self, request: "ModelRequest", *, stream: Literal[True] = True
    ) -> AsyncGenerator["ModelStreamCompletions", None]: ...

    @abstractmethod
    async def ask(
        self, request: "ModelRequest", *, stream: bool = False
    ) -> Union["ModelCompletions", AsyncGenerator["ModelStreamCompletions", None]]:
        """
        模型交互询问

        :param request: 模型调用请求体
        :param stream: 是否开启流式对话

        :return: 模型输出体
        """
        pass


@dataclass
class ModelRequest:
    """
    模型调用请求
    """

    prompt: str
    history: List[Message] = field(default_factory=list)
    resources: List[Resource] = field(default_factory=list)
    tools: Optional[List[dict]] = field(default_factory=list)
    system: Optional[str] = None


@dataclass
class ModelCompletions:
    """
    模型输出
    """

    text: str = ""
    usage: int = -1
    resources: List[Resource] = field(default_factory=list)
    succeed: Optional[bool] = True


@dataclass
class ModelStreamCompletions:
    """
    模型流式输出
    """

    chunk: str = ""
    usage: int = -1
    resources: Optional[List[Resource]] = field(default_factory=list)
    succeed: Optional[bool] = True


class FunctionCallRequest(BaseModel):
    """
    模型 FunctionCall 请求
    """

    func: str
    """函数名称"""
    arguments: dict[str, str] | None = None
    """函数参数"""



import inspect
from pathlib import Path
from typing import Any, Callable


def path_to_module_name(module_path: Path, base_path: Path) -> str:
    """
    动态计算模块名，基于明确的基准路径
    """
    try:
        rel_path = module_path.resolve().relative_to(base_path.resolve())
    except ValueError:
        # 处理绝对路径与相对路径的兼容性问题
        rel_path = module_path.resolve()

    if rel_path.stem == "__init__":
        parts = rel_path.parts[:-1]
    else:
        parts = rel_path.parts

    # 过滤空字符串和无效部分
    module_names = [p for p in parts if p not in ("", ".", "..")]
    return ".".join(module_names)


def is_coroutine_callable(call: Callable[..., Any]) -> bool:
    """
    检查 call 是否是一个 callable 协程函数
    """
    if inspect.isroutine(call):
        return inspect.iscoroutinefunction(call)
    if inspect.isclass(call):
        return False
    func_ = getattr(call, "__call__", None)
    return inspect.iscoroutinefunction(func_)



from nonebot import get_plugin_config, logger
from nonebot.adapters import Bot, Event
from nonebot.exception import IgnoredException
from nonebot.message import run_preprocessor
from nonebot_plugin_session import SessionIdType, extract_session
from pydantic import BaseModel

from muicebot.plugin import PluginMetadata


class ScopeConfig(BaseModel):
    blacklist: list[str] = []
    whitelist: list[str] = []


class Config(BaseModel):
    access_control: ScopeConfig = ScopeConfig()


__metadata__ = PluginMetadata(
    name="blacklist_whitelist_checker",
    description="黑白名单检测",
    usage="在插件配置中填写响应配置后即可",
    config=Config,
)

plugin_config = get_plugin_config(Config).access_control  # 获取插件配置

_BLACKLIST = plugin_config.blacklist
_WHITELIST = plugin_config.whitelist
_MODE = "white" if _WHITELIST else "black"


@run_preprocessor
async def access_control(bot: Bot, event: Event):
    session = extract_session(bot, event)
    group_id = session.get_id(SessionIdType.GROUP)
    user_id = session.get_id(SessionIdType.USER)
    level = session.level

    if _MODE == "black":
        if user_id in _BLACKLIST:
            msg = f"User {user_id} is in the blacklist"
            logger.warning(msg)
            raise IgnoredException(msg)

        elif group_id in _BLACKLIST:
            msg = f"Group {group_id} is in the blacklist"
            logger.warning(msg)
            raise IgnoredException(msg)

    if _MODE == "white":
        if level >= 2 and group_id not in _WHITELIST:  # 白名单只对群组生效
            msg = f"Group {group_id} is not in the whitelist"
            logger.warning(msg)
            raise IgnoredException(msg)



from nonebot.adapters import Bot, Event

from muicebot.plugin import PluginMetadata
from muicebot.plugin.func_call import on_function_call
from muicebot.utils.utils import get_username as get_username_

__metadata__ = PluginMetadata(
    name="muicebot-plugin-username", description="获取用户名的插件", usage="直接调用，返回当前对话的用户名"
)


@on_function_call(description="获取当前对话的用户名字")
async def get_username(bot: Bot, event: Event) -> str:
    user_id = event.get_user_id()
    return await get_username_(user_id)



import random

from apscheduler.schedulers.asyncio import AsyncIOScheduler
from apscheduler.triggers.cron import CronTrigger
from apscheduler.triggers.interval import IntervalTrigger
from nonebot import get_bot, logger
from nonebot_plugin_alconna.uniseg import Target, UniMessage

from .config import get_schedule_configs
from .models import Message
from .muice import Muice


async def send_message(target_id: str, message: str, probability: float = 1):
    """
    定时任务：发送信息

    :param target_id: 目标id；若为群聊则为 group_id 或者 channel_id，若为私聊则为 user_id
    :param message: 要发送的消息
    :param probability: 发送几率
    """
    if not (random.random() < probability):
        return

    logger.info(f"定时任务: send_message: {message}")

    target = Target(target_id)
    await UniMessage(message).send(target=target, bot=get_bot())


async def model_ask(muice_app: Muice, target_id: str, prompt: str, probability: float = 1):
    """
    定时任务：向模型发送消息

    :param muice_app: 沐雪核心类，用于与大语言模型交互
    :param target_id: 目标id；若为群聊则为 group_id 或者 channel_id，若为私聊则为 user_id
    :param prompt: 模型提示词
    :param probability: 发送几率
    """
    if not (random.random() < probability):
        return

    logger.info(f"定时任务: model_ask: {prompt}")

    if muice_app.model and muice_app.model.is_running:
        message = Message(message=prompt, userid=f"(bot_ask){target_id}")
        response = await muice_app.ask(message, enable_history=False, enable_plugins=False)

        target = Target(target_id)
        await UniMessage(response.text).send(target=target, bot=get_bot())


def setup_scheduler(muice: Muice) -> AsyncIOScheduler:
    """
    设置任务调度器

    :param muice: 沐雪核心类，用于与大语言模型交互
    """
    jobs = get_schedule_configs()
    scheduler = AsyncIOScheduler()

    for job in jobs:
        job_id = job.id
        job_type = "send_message" if job.say else "model_ask"
        trigger_type = job.trigger
        trigger_args = job.args

        # 解析触发器
        if trigger_type == "cron":
            trigger = CronTrigger(**trigger_args)

        elif trigger_type == "interval":
            trigger = IntervalTrigger(**trigger_args)

        else:
            logger.error(f"未知的触发器类型: {trigger_type}")
            continue

        # 添加任务
        if job_type == "send_message":
            scheduler.add_job(
                send_message,
                trigger,
                id=job_id,
                replace_existing=True,
                args=[job.target, job.say, job.probability],
            )
        else:
            scheduler.add_job(
                model_ask,
                trigger,
                id=job_id,
                replace_existing=True,
                args=[muice, job.target, job.ask, job.probability],
            )

        logger.success(f"已注册定时任务: {job_id}")

    if jobs:
        scheduler.start()
    return scheduler



import asyncio
import shutil
from pathlib import Path
from typing import Optional

import aiohttp
from nonebot import logger

from muicebot.plugin import load_plugin

from .config import config
from .models import PluginInfo
from .register import load_json_record, register_plugin, unregister_plugin

PLUGIN_DIR = Path("plugins/store")
PLUGIN_DIR.mkdir(parents=True, exist_ok=True)


async def get_index() -> Optional[dict[str, PluginInfo]]:
    """
    获取插件索引
    """
    logger.info("获取插件索引文件...")
    try:
        async with aiohttp.ClientSession() as session:
            async with session.get(config.store_index) as response:
                response.raise_for_status()
                return await response.json(content_type=None)
    except aiohttp.ClientError as e:
        logger.error(f"获取插件索引失败: {e}")
    except Exception as e:
        logger.exception(f"解析插件索引时发生意外错误: {e}")
    return {}


async def get_plugin_commit(plugin: str) -> str:
    """
    获取插件 commit hash
    """
    plugin_path = PLUGIN_DIR / plugin

    process = await asyncio.create_subprocess_exec(
        "git",
        "log",
        '--pretty=format:"%h"',
        "-1",
        cwd=plugin_path,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    stdout, stderr = await process.communicate()

    return stdout.decode().strip()


def load_store_plugin():
    """
    加载商店插件
    """
    logger.info("加载商店插件...")
    plugins = load_json_record()

    for plugin, info in plugins.items():
        if not Path(PLUGIN_DIR / plugin).exists():
            continue

        module_path = PLUGIN_DIR / plugin / info["module"]
        load_plugin(module_path)


async def install_dependencies(path: Path) -> bool:
    """
    安装插件依赖

    :return: 依赖安装状态
    """
    logger.info("安装插件依赖...")

    if (path / "pyproject.toml").exists():
        cmd = ["python", "-m", "pip", "install", "."]
    elif (path / "requirements.txt").exists():
        cmd = ["python", "-m", "pip", "install", "-r", "requirements.txt"]
    else:
        return True

    proc = await asyncio.create_subprocess_exec(
        *cmd,
        cwd=str(path),
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    stdout, stderr = await proc.communicate()

    if proc.returncode == 0:
        return True
    else:
        logger.error("插件依赖安装失败!")
        logger.error(stderr)
        return False


async def get_installed_plugins_info() -> str:
    """
    获得已安装插件信息
    """
    plugins = load_json_record()
    plugins_info = []
    for plugin, info in plugins.items():
        plugins_info.append(f"{plugin}: {info['name']} {info['commit']}")
    return "\n".join(plugins_info) or "本地还未安装商店插件~"


async def install_plugin(plugin: str) -> str:
    """
    通过 git clone 安装指定插件
    """
    if not (index := await get_index()):
        return "❌ 无法获取插件索引文件，请检查控制台日志"

    if plugin not in index:
        return f"❌ 插件 {plugin} 不存在于索引中！请检查插件名称是否正确"

    repo_url = index[plugin]["repo"]
    module = index[plugin]["module"]
    name = index[plugin]["name"]
    plugin_path = PLUGIN_DIR / plugin

    if plugin_path.exists():
        return f"⚠️ 插件 {plugin} 已存在，无需安装。"

    logger.info(f"获取插件: {repo_url}")
    try:
        process = await asyncio.create_subprocess_exec(
            "git",
            "clone",
            repo_url,
            str(plugin_path),
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        stdout, stderr = await process.communicate()

        if process.returncode != 0:
            return f"❌ 安装失败：{stderr.decode().strip()}"

        if not await install_dependencies(plugin_path):
            return "❌ 插件依赖安装失败！请检查控制台输出"

    except FileNotFoundError:
        return "❌ 请确保已安装 Git 并配置到 PATH。"

    load_plugin(plugin_path / module)

    commit = await get_plugin_commit(plugin)

    register_plugin(plugin, commit, name, module)

    return f"✅ 插件 {plugin} 安装成功！"


async def update_plugin(plugin: str) -> str:
    """
    更新指定插件
    """
    plugin_path = PLUGIN_DIR / plugin

    if not plugin_path.exists():
        return f"❌ 插件 {plugin} 不存在！"

    logger.info(f"更新插件: {plugin}")
    try:
        process = await asyncio.create_subprocess_exec(
            "git",
            "pull",
            cwd=plugin_path,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        stdout, stderr = await process.communicate()

        if process.returncode != 0:
            return f"❌ 插件更新失败：{stderr.decode().strip()}"

    except FileNotFoundError:
        return "❌ 请确保已安装 Git 并配置到 PATH。"

    await install_dependencies(plugin_path)

    info = load_json_record()[plugin]
    commit = await get_plugin_commit(plugin)
    register_plugin(plugin, commit, info["name"], info["module"])

    return f"✅ 插件 {plugin} 更新成功！重启后生效"


async def uninstall_plugin(plugin: str) -> str:
    """
    卸载指定插件
    """
    plugin_path = PLUGIN_DIR / plugin

    if not plugin_path.exists():
        return f"❌ 插件 {plugin} 不存在！"

    logger.info(f"卸载插件: {plugin}")

    unregister_plugin(plugin)

    try:
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, shutil.rmtree, plugin_path)
    except PermissionError:
        return f"❌ 插件 {plugin} 移除失败，请尝试手动移除"

    return f"✅ 插件 {plugin} 移除成功！重启后生效"



from ._dependencies import MODEL_DEPENDENCY_MAP, get_missing_dependencies
from ._types import (
    BasicModel,
    ModelCompletions,
    ModelConfig,
    ModelRequest,
    ModelStreamCompletions,
)

__all__ = [
    "BasicModel",
    "ModelConfig",
    "ModelRequest",
    "ModelCompletions",
    "ModelStreamCompletions",
    "MODEL_DEPENDENCY_MAP",
    "get_missing_dependencies",
]



from muicebot.plugin import PluginMetadata
from muicebot.plugin.func_call import on_function_call

__metadata__ = PluginMetadata(
    name="muicebot-plugin-time", description="时间插件", usage="直接调用，返回 %Y-%m-%d %H:%M:%S 格式的当前时间"
)


@on_function_call(
    description="获取当前时间",
)
async def get_current_time() -> str:
    """
    获取当前时间
    """
    import datetime

    current_time = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    return current_time



from pathlib import Path

import yaml
from jinja2 import Environment, FileSystemLoader
from jinja2.exceptions import TemplateNotFound
from nonebot import logger

from .model import PromptTemplatesConfig, PromptTemplatesData

SEARCH_PATH = ["./templates", Path(__file__).parent.parent / "builtin_templates"]

TEMPLATES_CONFIG_PATH = "./configs/templates.yml"


def load_templates_config() -> dict:
    """
    获取模板配置
    """
    try:
        with open(TEMPLATES_CONFIG_PATH, "r", encoding="utf-8") as f:
            return yaml.safe_load(f) or {}
    except (FileNotFoundError, yaml.YAMLError):
        return {}


def load_templates_data(userid: str, is_private: bool = False) -> PromptTemplatesData:
    """
    获取模板数据
    """
    config = load_templates_config()
    templates_config = PromptTemplatesConfig(**config)
    return PromptTemplatesData.from_config(templates_config, userid=userid, is_private=is_private)


def generate_prompt_from_template(template_name: str, userid: str, is_private: bool = False) -> str:
    """
    获取提示词
    """
    env = Environment(loader=FileSystemLoader(SEARCH_PATH))

    if not template_name.endswith((".j2", ".jinja2")):
        template_name += ".jinja2"
    try:
        template = env.get_template(template_name)
    except TemplateNotFound:
        logger.error(f"模板文件 {template_name} 未找到!")
        return ""

    templates_data = load_templates_data(userid, is_private)
    prompt = template.render(templates_data.model_dump())

    return prompt
